/****************************************
 *  COPYRIGHT (C) 2021-2023
 *  Holger Graf
 ****************************************/
package datatransfer;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;

import com.fasterxml.jackson.jr.ob.JSON;

import datatransfer.json.InfoBean;
import datatransfer.json.RequestBean;
import datatransfer.json.ResponseBean;
import net.freeutils.httpserver.HTTPServer;
import net.freeutils.httpserver.HTTPServer.ContextHandler;
import net.freeutils.httpserver.HTTPServer.Request;
import net.freeutils.httpserver.HTTPServer.Response;
import net.freeutils.httpserver.HTTPServer.VirtualHost;

/**
 * @author graf
 *
 */
public abstract class DataServer {

    private HTTPServer server = null;
    private static final int apiVersion = 1;
    private static final int[] serverPorts = { 4000, 4001, 4002, 4003, 4004, 4005, 4006, 4007, 4008, 4009};

    private static final SimpleDateFormat time = new SimpleDateFormat( "dd.MM.yyyy HH:mm:ss.s" );
    private int port = 0;

    public DataServer() {
    }

    public int startServer() {
        port = 0;
        boolean retry;
        if(serverPorts.length > 0) {
            int idx = 0;
            port = serverPorts[idx];
            server = setupServer(port);
            do {
                try {
                    server.start();
                    retry = false;
                }
                catch(BindException ex) {
                    idx++;
                    if(idx < serverPorts.length) {
                        port = serverPorts[idx];
                        server.setPort(port);
                        retry = true;
                    }
                    else {
                        port = 0;
                        retry = false;
                    }
                }
                catch(IOException ex) {
                    port = 0;
                    retry = false;
                }
            } while(retry);
        }
        String msg;
        if(port != 0) {
            msg = "HTTP server listening on port " + port;
        }
        else {
            msg = "HTTP server not started";
        }
        System.out.println(msg);
        return port;
    }

    public void stopServer() {
        port = 0;
        if(server != null) {
            server.stop();
            System.out.println("HTTP server stopped");
        }
    }

    public int getPort() {
        return port;
    }

    protected abstract String getIdentifier();
    protected abstract String getProductName();
    protected abstract String getProductVersion();
    protected abstract Integer getAccountVersionID();

    protected int serveRequest(final RequestType type, Request req, Response resp) throws IOException {
        Map<String, String> params = req.getParams();
        String origin = req.getHeaders().get("Origin");
        resp.getHeaders().add("Content-Type", "text/json");
        resp.getHeaders().add("Access-Control-Allow-Origin", origin);
        ResponseBean responseBean = new ResponseBean();
        responseBean.setStatus(0);
        responseBean.setName(getProductName());
        responseBean.setIdentifier(getIdentifier());
        responseBean.setVersion(getProductVersion());
        responseBean.setTversion(apiVersion);
        String json = params.get("json");
        if( json != null ) {
            try {
                updateResponseBean(responseBean, type, json);
            }
            catch(IOException ex) {
                // parse error
                responseBean.setStatus(1);
            }
        }
        resp.send(200,  JSON.std.asString(responseBean));
        return 0;
    }

    protected int failedUpdate(Request req, Response resp) throws IOException {
        Map<String, String> params = req.getParams();
        // log whatever is there
        String logDir = System.getProperty( "user.home" );
        String fileName = "si-connect-update.log";
        long now = System.currentTimeMillis();
        try(PrintWriter pw = new PrintWriter(new FileWriter(logDir + File.separator + fileName, true))) {
            pw.println(time.format(new Date(now)));
            System.err.println(time.format(new Date(now)));
            for(Map.Entry<String, String> entry : params.entrySet()) {
                pw.print(entry.getKey());
                System.err.print(entry.getKey());
                pw.print(": ");
                System.err.print(": ");
                pw.println(entry.getValue());
                System.err.println(entry.getValue());
            }
            pw.println("----------------------------------------");
            System.err.println("----------------------------------------");
        }
        resp.send(200, "Logged");
        return 0;
    }
    /**
     * Override to parse the JSON input data.
     * ResponseBean fields can be overwritten as needed. Most are pre-filled with sensible values.
     * It is safe to ignore request types you do not wish to react to.
     *
     * @param responseBean
     * @param type the request type
     * @param json JSON input data
     * @throws IOException when JSON data cannot be parsed successfully
     */
    protected void updateResponseBean(ResponseBean responseBean, RequestType type, String json) throws IOException  {
        switch(type) {
            case init:
                {
                    RequestBean requestBean = JSON.std.beanFrom(RequestBean.class, json);
                    int user = requestBean.getInfo().getUser().getUserID();
                    String k = requestBean.getK();
                    StringBuilder s = new StringBuilder();
                    s.append(getIdentifier())
                     .append(user)
                     .append(k);
                    responseBean.setC(sha256(s.toString()));
                    responseBean.setStatus(0);
                }
                break;
            default:
                break;
        }
    }

    protected HTTPServer setupServer(final int port) {
        HTTPServer server = new HTTPServer(port) {
            /**
             * Overridden to bind only to localhost
             * @see net.freeutils.httpserver.HTTPServer#createServerSocket()
             */
            @Override
            protected ServerSocket createServerSocket() throws IOException {
                ServerSocket serv = serverSocketFactory.createServerSocket();
                serv.setReuseAddress(true);
                serv.bind(new InetSocketAddress(InetAddress.getLoopbackAddress(), port));
                return serv;
            }
        };

        VirtualHost host = server.getVirtualHost(null);

        host.addContext(RequestType.init.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.init, req, resp);
            }
        }, "POST");

        host.addContext(RequestType.ping.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.ping, req, resp);
            }
        }, "POST");

        host.addContext(RequestType.failed.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return failedUpdate(req, resp);
            }
        }, "POST");

        host.addContext(RequestType.battlereport.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.battlereport, req, resp);
            }
        }, "POST");

        host.addContext(RequestType.galaxyOverview.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.galaxyOverview, req, resp);
            }
        }, "POST");

        host.addContext(RequestType.planetOverview.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.planetOverview, req, resp);
            }
        }, "POST");

        host.addContext(RequestType.spyreport.getEndpoint(), new ContextHandler() {
            public int serve(Request req, Response resp) throws IOException {
                return serveRequest(RequestType.spyreport, req, resp);
            }
        }, "POST");

        server.addVirtualHost(host);
        return server;
    }

    protected static String sha256(String input) {
        MessageDigest digest;
        try {
            digest = MessageDigest.getInstance("SHA-256");
        }
        catch(NoSuchAlgorithmException e) {
            return "";
        }
        byte[] hash = digest.digest(input.getBytes(StandardCharsets.UTF_8));
        return toHexString(hash);
    }

    protected static String toHexString(byte[] bytes) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < bytes.length; i++) {
            sb.append(Integer.toHexString((bytes[i] >>> 4) & 0xF)).append(Integer.toHexString(bytes[i] & 0xF));
        }
        return sb.toString();
    }

    protected boolean checkAccess(InfoBean infoBean) {
        boolean rv = false;
        // this is a hack that lets the beta test uni look like sirius
        int id = infoBean.getUniverse().getId() % 855;
        rv = (getAccountVersionID() == id);
        return rv;
    }
}
